/*

    File: ntfs_adv.c

    Copyright (C) 1998-2004 Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
 
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <ctype.h>
#include "types.h"
#include "common.h"
#include "intrf.h"
#include "ntfs.h"
#include "fnctdsk.h"
#include "lang.h"
#include "io_redir.h"

static int create_ntfs_boot_sector(t_param_disk *disk_car, t_partition *partition, const int interface, const unsigned int cluster_size, const uint64_t mft_lcn, const uint64_t mftmirr_lcn, const uint32_t mft_record_size, const uint32_t index_block_size);
static int dump_2ntfs_info(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2);
static int dump_ntfs_info(const struct ntfs_boot_sector *ntfs_header);
static int testdisk_ffs(int x);
static int read_mft_info(t_param_disk *disk_car, t_partition *partition, uint64_t mft_sector, const int debug, unsigned int *sectors_per_cluster, uint64_t *mft_lcn, uint64_t *mftmirr_lcn, unsigned int *mft_record_size);


static int create_ntfs_boot_sector(t_param_disk *disk_car, t_partition *partition, const int interface, const unsigned int cluster_size, const uint64_t mft_lcn, const uint64_t mftmirr_lcn, const uint32_t mft_record_size, const uint32_t index_block_size)
{
  unsigned char orgboot[DEFAULT_SECTOR_SIZE];
  unsigned char newboot[DEFAULT_SECTOR_SIZE];
  struct ntfs_boot_sector *org_ntfs_header=(struct ntfs_boot_sector *)&orgboot;
  struct ntfs_boot_sector *ntfs_header=(struct ntfs_boot_sector *)&newboot;
  int error=0;
  if(disk_car->read(disk_car,sizeof(orgboot), &orgboot, partition->part_offset)!=0)
  {
    ecrit_rapport("create_ntfs_boot_sector: Can't read boot sector.\n");
    memset(&orgboot,0,sizeof(orgboot));
  }
  if(cluster_size==0)
  {
    error=1;
  }
  if(error)
  {
    display_message("NTFS Bad extrapolation.\n");
    return 1;
  }
  memcpy(&newboot,&orgboot,sizeof(newboot));
  memcpy(ntfs_header->system_id,"NTFS    ",8);
  ntfs_header->sector_size[0]=disk_car->sector_size & 0xFF;
  ntfs_header->sector_size[1]=disk_car->sector_size>>8;
  ntfs_header->sectors_per_cluster=cluster_size/disk_car->sector_size;
  ntfs_header->reserved=0;
  ntfs_header->fats=0;
  ntfs_header->dir_entries[0]=0;
  ntfs_header->dir_entries[1]=0;
  ntfs_header->sectors[0]=0;
  ntfs_header->sectors[1]=0;
  ntfs_header->media=0xF8;
  ntfs_header->fat_length=0;
  ntfs_header->secs_track=le16(disk_car->CHS.sector);
  ntfs_header->heads=le16(disk_car->CHS.head+1);
  if(partition->status==STATUS_LOG)
    ntfs_header->hidden=63;
  else
    ntfs_header->hidden=partition->part_offset/disk_car->sector_size;
  ntfs_header->total_sect=0;
  ntfs_header->sectors_nbr=le64(partition->part_size/disk_car->sector_size-1);
  ntfs_header->mft_lcn=mft_lcn;
  ntfs_header->mftmirr_lcn=mftmirr_lcn;
  ntfs_header->clusters_per_mft_record=(mft_record_size >= cluster_size?mft_record_size / cluster_size:
      -(testdisk_ffs(mft_record_size) - 1));
  ntfs_header->clusters_per_index_record =(index_block_size >= cluster_size?index_block_size / cluster_size:
      -(testdisk_ffs(index_block_size) - 1));
  ntfs_header->reserved0[0]=0;
  ntfs_header->reserved0[1]=0;
  ntfs_header->reserved0[2]=0;
  ntfs_header->reserved1[0]=0;
  ntfs_header->reserved1[1]=0;
  ntfs_header->reserved1[2]=0;
  /*
  {
    uint32_t *u;
    uint32_t checksum;
    for (checksum = 0,u=(uint32_t*)ntfs_header; u < (uint32_t*)(&ntfs_header->checksum); u++)
      checksum += NTFS_GETU32(u);
    ntfs_header->checksum=checksum;
  }
  */
  ntfs_header->checksum=0;
  ntfs_header->marker=0xAA55;
  if(memcmp(newboot,orgboot,sizeof(newboot)))
  {
    ecrit_rapport("             New / Current boot sector\n");
    dump_2ntfs_rapport(ntfs_header,org_ntfs_header);
    ecrit_rapport("Extrapolated boot sector and current boot sector are different.\n");
  }
  else
  {
    ecrit_rapport("Extrapolated boot sector and current boot sector are identical.\n");
  }
  /* */
  if(interface)
  {
    struct MenuItem menuSaveBoot[]=
    {
      { 'D', "Dump", "Dump sector" },
      { 'L', "List", "List directories and files" },
      { 'W', "Write","Write boot"},
      { 'Q',"Quit","Quit this section"},
      { 0, NULL, NULL }
    };
    const char *options="DLQ";
    int do_write=0;
    int do_exit=0;
    do
    {
      aff_copy(stdscr);
      wmove(stdscr,4,0);
      wdoprintf(stdscr,"%s",disk_car->description(disk_car));
      mvwaddstr(stdscr,5,0,msg_PART_HEADER_LONG);
      wmove(stdscr,6,0);
      aff_part(stdscr,AFF_PART_ORDER,disk_car,partition);
      wmove(stdscr,8,0);
      if(memcmp(newboot,orgboot,sizeof(newboot)))	/* Only compare the first sector */
      {
	options="DLWQ";
	dump_2ntfs_info(ntfs_header, org_ntfs_header);
	dump_2ntfs_rapport(ntfs_header, org_ntfs_header);
	wdoprintf(stdscr,"Extrapolated boot sector and current boot sector are different.\n");
	if(error)
	  ecrit_rapport("Warning: Extrapolated boot sector have incorrect values.\n");
      }
      else
      {
	dump_ntfs_info(ntfs_header);
	dump_ntfs_rapport(ntfs_header);
	wdoprintf(stdscr,"Extrapolated boot sector and current boot sector are identical.\n");
      }
      switch(toupper(wmenuSelect(stdscr,INTER_DUMP_Y, INTER_DUMP_X, menuSaveBoot,8,options,MENU_HORIZ | MENU_BUTTON, 1)))
      {
	case 'W':
	  if(strchr(options,'W')!=NULL)
	    do_write=1;
	  break;
	case 'D':
	  if(strchr(options,'D')!=NULL)
	  {
	    dump2(stdscr,newboot,orgboot, sizeof(newboot));
	  }
	  break;
	case 'L':
	  io_redir_add_redir(disk_car,partition->part_offset,sizeof(newboot),0,newboot);
	  dir_partition_ntfs(stdscr,disk_car, partition, 0);
	  io_redir_del_redir(disk_car,partition->part_offset);
	  break;
	case 'Q':
	  do_exit=1;
	  break;
      }
    } while(do_write==0 && do_exit==0);
    if(do_write!=0 && ask_confirmation("Write new NTFS boot sector, confirm ? (Y/N)")!=0)
    {
      ecrit_rapport("Write new boot!\n");
      /* Write boot sector and backup boot sector */
      if(disk_car->write(disk_car,sizeof(newboot), &newboot, partition->part_offset))
      {
	display_message("Write error: Can't write new NTFS boot sector\n");
      }
      if(disk_car->write(disk_car,sizeof(newboot), &newboot, partition->part_offset+partition->part_size-disk_car->sector_size)!=0)
      {
	display_message("Write error: Can't write new NTFS backup boot sector\n");
      }
    }
    else
      ecrit_rapport("Don't write new boot!\n");
  }
  return 0;
}

static int read_mft_info(t_param_disk *disk_car, t_partition *partition, uint64_t mft_sector, const int debug, unsigned int *sectors_per_cluster, uint64_t *mft_lcn, uint64_t *mftmirr_lcn, unsigned int *mft_record_size)
{
  char buffer[8*DEFAULT_SECTOR_SIZE];
  const char *attr=buffer;
  if(disk_car->read(disk_car,sizeof(buffer), &buffer, partition->part_offset+(uint64_t)mft_sector*disk_car->sector_size)!=0)
  {
    display_message("NTFS: Can't read mft_sector\n");
    return 1;
  }
  *mft_lcn=ntfs_get_attr(attr,0x80,partition,buffer+8*DEFAULT_SECTOR_SIZE,debug,0,NULL);
  *mft_record_size=NTFS_GETU32(attr + 0x1C);
  if(*mft_record_size==0)
  {
    if(debug>0)
      ecrit_rapport("read_mft_info failed: mft_record_size=0\n");
    return 2;
  }
  attr+= NTFS_GETU32(attr + 0x1C);
  *mftmirr_lcn=ntfs_get_attr(attr,0x80,partition,buffer+8*DEFAULT_SECTOR_SIZE,debug,0,NULL);
  /* Try to divide by the biggest number first */
  if(*mft_lcn<*mftmirr_lcn)
  {
    if(*mftmirr_lcn>0 && mft_sector%(*mftmirr_lcn)==0)
    {
      *sectors_per_cluster=mft_sector/(*mftmirr_lcn);
      return 0;
    }
    if(*mft_lcn>0 && mft_sector%(*mft_lcn)==0)
    {
      *sectors_per_cluster=mft_sector/(*mft_lcn);
      return 0;
    }
  }
  else
  {
    if(*mft_lcn>0 && mft_sector%(*mft_lcn)==0)
    {
      *sectors_per_cluster=mft_sector/(*mft_lcn);
      return 0;
    }
    if(*mftmirr_lcn>0 && mft_sector%(*mftmirr_lcn)==0)
    {
      *sectors_per_cluster=mft_sector/(*mftmirr_lcn);
      return 0;
    }
  }
  if(debug>0)
  {
    ecrit_rapport("read_mft_info failed\n");
    ecrit_rapport("ntfs_find_mft: sectors_per_cluster %u\n",*sectors_per_cluster);
    ecrit_rapport("ntfs_find_mft: mft_lcn             %lu\n",(long unsigned int)*mft_lcn);
    ecrit_rapport("ntfs_find_mft: mftmirr_lcn         %lu\n",(long unsigned int)*mftmirr_lcn);
    ecrit_rapport("ntfs_find_mft: mft_record_size     %u\n",*mft_record_size);
    ecrit_rapport("\n");
  }
  return 3;
}

int rebuild_NTFS_BS(t_param_disk *disk_car, t_partition *partition, const int debug, const int dump_ind,const int interface)
{
  uint64_t sector;
  char buffer[8*DEFAULT_SECTOR_SIZE];
  int ind_stop=FALSE;
  unsigned int sectors_per_cluster=0;
  uint64_t mft_lcn;
  uint64_t mftmirr_lcn;
  unsigned int mft_record_size=1024;
  ecrit_rapport("rebuild_NTFS_BS\n");
  if(interface)
  {
    aff_copy(stdscr);
    wmove(stdscr,4,0);
    wdoprintf(stdscr,"%s",disk_car->description(disk_car));
    mvwaddstr(stdscr,5,0,msg_PART_HEADER_LONG);
    wmove(stdscr,6,0);
    aff_part(stdscr,AFF_PART_ORDER,disk_car,partition);
  }
  /* try to find MFT Backup first */
  for(sector=(partition->part_size/disk_car->sector_size/2-20>0?partition->part_size/disk_car->sector_size/2-20:1);(sector<partition->part_size/disk_car->sector_size)&&(sector<=partition->part_size/disk_car->sector_size/2+20)&&(ind_stop==FALSE);sector++)
  {
    if(disk_car->read(disk_car,2*DEFAULT_SECTOR_SIZE, &buffer, partition->part_offset+sector*(uint64_t)disk_car->sector_size)==0)
    {
      if(memcmp(buffer,"FILE",4)==0 && (NTFS_GETU16(buffer+ 0x14)%8==0) && (NTFS_GETU16(buffer+ 0x14)>=42)
	  &&(NTFS_GETU16(buffer+22)==1))	/* MFT_RECORD_IN_USE */
      {
	int res;
	res=ntfs_get_attr(buffer,0x30,partition,buffer+2*DEFAULT_SECTOR_SIZE,debug,0,"$MFT");
	if(res==1)
	{
	  ecrit_rapport("mft at %lu, seq=%u, main=%u res=%d\n",(long unsigned)sector,NTFS_GETU8(buffer+0x10),(unsigned int)NTFS_GETU32(buffer+0x20),res);
	  if(read_mft_info(disk_car, partition, sector, debug, &sectors_per_cluster, &mft_lcn, &mftmirr_lcn, &mft_record_size)==0)
	  {
	    ecrit_rapport("ntfs_find_mft: mft_lcn             %lu\n",(long unsigned int)mft_lcn);
	    ecrit_rapport("ntfs_find_mft: mftmirr_lcn         %lu\n",(long unsigned int)mftmirr_lcn);
	    ind_stop=TRUE;
	  }
	}
      }
    }
  }
  for(sector=1;(sector<partition->part_size/disk_car->sector_size)&&(ind_stop==FALSE);sector++)
  {
    if((interface!=0) &&(sector%200)==0)
    {
      wmove(stdscr,9,0);
      wclrtoeol(stdscr);
      wdoprintf(stdscr,"Search mft %10lu/%lu",(long unsigned)sector,(long unsigned)(partition->part_size/disk_car->sector_size));
      wrefresh(stdscr);
      ind_stop|=check_enter_or_s(stdscr);
    }
    if(disk_car->read(disk_car,2*DEFAULT_SECTOR_SIZE, &buffer, partition->part_offset+sector*(uint64_t)disk_car->sector_size)==0)
    {
      if(memcmp(buffer,"FILE",4)==0 && (NTFS_GETU16(buffer+ 0x14)%8==0) && (NTFS_GETU16(buffer+ 0x14)>=42))
      {
	int res;
	res=ntfs_get_attr(buffer,0x30,partition,buffer+2*DEFAULT_SECTOR_SIZE,debug,0,"$MFT");
	if(res==1)
	{
	  ecrit_rapport("mft at %lu, seq=%u, main=%u res=%d\n",(long unsigned)sector,NTFS_GETU8(buffer+0x10),(unsigned int)NTFS_GETU32(buffer+0x20),res);
	  if(read_mft_info(disk_car, partition, sector, debug, &sectors_per_cluster, &mft_lcn, &mftmirr_lcn, &mft_record_size)==0)
	  {
	    ecrit_rapport("ntfs_find_mft: mft_lcn             %lu\n",(long unsigned int)mft_lcn);
	    ecrit_rapport("ntfs_find_mft: mftmirr_lcn         %lu\n",(long unsigned int)mftmirr_lcn);
	    ind_stop=TRUE;
	  }
	}
      }
    }
  }
  /* TODO read_mft_info(partition,sector,*sectors_per_cluster,*mft_lcn,*mftmirr_lcn,*mft_record_size); */
  if(sectors_per_cluster>0 && mft_record_size>0)
  {
    unsigned int index_block_size=4096;
    /* Read "root directory" in MFT */
    if(disk_car->read(disk_car,mft_record_size, &buffer, partition->part_offset+(uint64_t)mft_lcn*sectors_per_cluster*disk_car->sector_size+5*(uint64_t)mft_record_size)!=0)
    {
      display_message("NTFS Can't read \"root directory\" in MFT\n");
      return 1;
    }
    index_block_size=ntfs_get_attr(buffer,0x90,partition,buffer+mft_record_size,debug,0,NULL);
    if(index_block_size%512!=0)
      index_block_size=4096;
    ecrit_rapport("ntfs_find_mft: sectors_per_cluster %u\n",sectors_per_cluster);
    ecrit_rapport("ntfs_find_mft: mft_lcn             %lu\n",(long unsigned int)mft_lcn);
    ecrit_rapport("ntfs_find_mft: mftmirr_lcn         %lu\n",(long unsigned int)mftmirr_lcn);
    ecrit_rapport("ntfs_find_mft: mft_record_size     %u\n",mft_record_size);
    ecrit_rapport("ntfs_find_mft: index_block_size    %u\n",index_block_size);
    create_ntfs_boot_sector(disk_car,partition, interface, sectors_per_cluster*disk_car->sector_size, mft_lcn, mftmirr_lcn, mft_record_size, index_block_size);
  }
  else
  {
    ecrit_rapport("Failed to rebuild NTFS boot sector.\n");
  }
  return 0;
}

static int dump_ntfs_info(const struct ntfs_boot_sector *ntfs_header)
{
  wdoprintf(stdscr,"sectors_per_cluster       %u\n",ntfs_header->sectors_per_cluster);
  wdoprintf(stdscr,"mft_lcn                   %lu\n",(long unsigned int)ntfs_header->mft_lcn);
  wdoprintf(stdscr,"mftmirr_lcn               %lu\n",(long unsigned int)ntfs_header->mftmirr_lcn);
  wdoprintf(stdscr,"clusters_per_mft_record   %d\n",ntfs_header->clusters_per_mft_record);
  wdoprintf(stdscr,"clusters_per_index_record %d\n",ntfs_header->clusters_per_index_record);
  return 0;
}

static int testdisk_ffs(int x)
{
  int r = 1;

  if (!x)
    return 0;
  if (!(x & 0xffff)) {
    x >>= 16;
    r += 16;
  }
  if (!(x & 0xff)) {
    x >>= 8;
    r += 8;
  }
  if (!(x & 0xf)) {
    x >>= 4;
    r += 4;
  }
  if (!(x & 3)) {
    x >>= 2;
    r += 2;
  }
  if (!(x & 1)) {
    x >>= 1;
    r += 1;
  }
  return r;
}

static int dump_2ntfs_info(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2)
{
  wdoprintf(stdscr,"sectors_per_cluster       %u %u\n",nh1->sectors_per_cluster,nh2->sectors_per_cluster);
  wdoprintf(stdscr,"mft_lcn                   %lu %lu\n",(long unsigned int)nh1->mft_lcn,(long unsigned int)nh2->mft_lcn);
  wdoprintf(stdscr,"mftmirr_lcn               %lu %lu\n",(long unsigned int)nh1->mftmirr_lcn,(long unsigned int)nh2->mftmirr_lcn);
  wdoprintf(stdscr,"clusters_per_mft_record   %d %d\n",nh1->clusters_per_mft_record,nh2->clusters_per_mft_record);
  wdoprintf(stdscr,"clusters_per_index_record %d %d\n",nh1->clusters_per_index_record,nh2->clusters_per_index_record);
  return 0;
}

int dump_2ntfs_rapport(const struct ntfs_boot_sector *nh1, const struct ntfs_boot_sector *nh2)
{
  ecrit_rapport("sectors_per_cluster       %u %u\n",nh1->sectors_per_cluster,nh2->sectors_per_cluster);
  ecrit_rapport("mft_lcn                   %lu %lu\n",(long unsigned int)nh1->mft_lcn,(long unsigned int)nh2->mft_lcn);
  ecrit_rapport("mftmirr_lcn               %lu %lu\n",(long unsigned int)nh1->mftmirr_lcn,(long unsigned int)nh2->mftmirr_lcn);
  ecrit_rapport("clusters_per_mft_record   %d %d\n",nh1->clusters_per_mft_record,nh2->clusters_per_mft_record);
  ecrit_rapport("clusters_per_index_record %d %d\n",nh1->clusters_per_index_record,nh2->clusters_per_index_record);
  return 0;
}


